;******************************************************************************
;*                                                                            *
;*                            License and T&C                                 *
;*                                                                            *
;*        GR8BIT storage subsystem microcontroller microcode version 1.0      *
;*             This software is copyleft 2011 by Eugeny Brychkov              *
;*                                                                            *
;* You can use and modify its source and executable, given two conditions are *
;* met: (a) you keep this Copyleft statement in the source code, and you do   *
;* not distribute software and resulting executable on the commercial basis   *
;* and on form the purposes, not related to GR8BIT platform, without written  *
;* permission of Eugeny Brychkov (eugeny_brychkov@hotmail.com)                *
;*                                                                            *
;******************************************************************************

; version 1.0 (Nov 28,2011)		- initial release
; version 1.0a (Jan 26,2012)	- Storage subsystem ROM was released, this code was intact
; version 1.0b (Mar 14, 2012)	- added support for 720K media, fixed some bugs regarding
;								  handling of i/o errors and disk removals, improved
;								  disk change handling


; GR8BIT storage subsystem requires at least single wait state at the host speed
; 3.579545 MHz for microcontroller access and for CPU /M1 cycle.

; The principle of operation of the microcontroller
; There's specific handshaking protocol to communicate with host which purpose
; is to provide reliable operation of the GR8BIT storage subsystem. It works through
; host ports 0xA4 and 0xA4, former is a control port, latter is data/status port.
; Writing to 0xA4 port cancels the stream of data and returns microcontroller into
; ready state. Every command has its own input and output data.

; IMPORTANT TECHNICAL INFORMATION
; FDD Controller works in Base mode
; instruction cycle: 125 ns (4*FOSC) at 32MHz
; /WR width: 60 ns min
; control signals setup relative to /WR: 0 ns min (no max)
; data set up to /WR high: 80 ns min (no max)
; data hold time from /WR high: 0 ns

; GAP lengths can be found in SMC37C65 data sheet
; Info on DSPEC values can be found here http://wiki.osdev.org/Floppy_Disk_Controller, also http://www.isdaman.com/alsos/hardware/fdc/floppy.htm
; 1.44M diskette: GPL1=0x1B, GPL2=0x54
; 720K diskette: GPL1=2Ah, GPL2=50h ?????
GPL1_144	equ		0x1B		; gap length when reading
GPL2_144	equ		0x54		; gap length when formatting
DSPEC1_144	equ		0x8F		; 8=step rate 8ms, F=head unload delay (240ms)
DSPEC2_144	equ		0x0B		; 5=10ms head load time, 1=no DMA
DSPD_144	equ		0x0			; speed 500kB/s
SpT_144		equ		0x12		; sectors per track (18)
; 720K
GPL1_720	equ	0x2A
GPL2_720	equ	0x50
DSPEC1_720	equ	0xCF		; C=step rate 8ms, F=head unload time (480ms)
DSPEC2_720	equ	0x07		; 5=12ms head load time, 1=no DMA
DSPD_720	equ	0x02		; speed 250kB/s (CR1=1, CR0=0, DRV=x)
SpT_720		equ	0x09		; sectors per track (9)

	list		p=16f1939			; list directive to define processor
	#include	<P16F1939.inc>		; processor specific variable definitions

	
    __CONFIG _CONFIG1, _FOSC_INTOSC & _WDTE_OFF & _PWRTE_OFF & _MCLRE_OFF & _CP_OFF & _CPD_OFF & _BOREN_OFF & _CLKOUTEN_OFF & _IESO_OFF & _FCMEN_OFF
    __CONFIG _CONFIG2, _WRT_OFF & _PLLEN_ON & _STVREN_OFF & _BORV_19 & _LVP_OFF


; Command buffer memory organization
;	output: bank 0, sequence of the bytes (or words) to send to controller (offset 0x50-0x5F)
;	intput: bank 0, sequence of the bytes (or words) read from the controller (offset 0x60-0x6F)

; Data buffer memory organization
;	bank 2, fist 64 (0x40) bytes, (offset 0x20-0x5F)
;	bank 3, second 64 (0x40) bytes, (offset 0x20-0x5F)
;	bank 4, third 64 (0x40) bytes, (offset 0x20-0x5F)
;	bank 5, last (fourth) 64 (0x40) bytes, (offset 0x20-0x5F)
;	...
;	bank 9, last (eighth) 64 (0x40) bytes, (offset 0x20-0x5F)
;	----- total 512 bytes for the sector


CTRSTAT	equ	0x20		; controller status - free/busy
TMP		equ	0x21
MAINSTAT	equ	0x22		; last read master status register of FDD
SENSEST	equ	0x23		; last read drive sense status (status reg 3 - ST3)
FDCQPOS	equ	0x24
QTEMP	equ	0x25
CMDRUN	equ	0x26		; not zero if command is running
FDCRUNPOS	equ	0x27	; position of the pointer when writing sequence to the FDD controller
RESULTN		equ	0x28	; number of result bytes to read from FDD

FMTTMP1	equ	0x29
FMTTMP2	equ	0x2A
FMTTMP3	equ	0x2B

DEBUGN		equ	0x2C

TIMOUT1		equ	0x2D	; used within timout counting code
TIMOUT2		equ	0x2E	; used within timout counting code
FDCERR		equ	0x2F	; not 0 if error
DATAVAL		equ	0x30
MOTATO		equ	0x31	; motor drive A timeout (0=no timeout)
MOTBTO		equ	0x32	; motor drive B timeout (0=no timeout)
DORVAL		equ	0x33	; current value of the DOR
CURDRV		equ	0x34	; Current disk drive (0=A, 1=B)

RECLENL	equ	0x35
RECLENH	equ	0x36

;DSKCHGFLG	equ	0x37		; disk change flag saved to be returned during next DSKCHG call
HOSTCMD	equ	0x38
TEMPREQ	equ	0x39		; temporary value for forming request to FDC
TEMPRE1	equ	0x3A		; temporary value for forming request to FDC
COUVAL	equ	0x3B		; command/MC-host interface counter value
DRINUM	equ	0x3C		; number of physical drives connected to the FDC
DRVPRM	equ	0x3D		; drive and head parameter in FDC format (CURDRV is just drive number 0 or 1)
CYLPRM	equ	0x3E		; cylinder parameter in FDC format
SECPRM	equ	0x3F		; sector parameter in FDC format
FMTPRM	equ	0x40
FMTPR1	equ	0x41		; three variables for format's writing control records
FMTPR2	equ	0x42
FMTPR3	equ	0x43
CURHOSTDRV	equ	0x44

SPDCOU	equ	0x45		; Counter for speed retry
DSK0SPD	equ	0x46		; DSK*SPD contain speed current disk/disk drive is using
DSK1SPD	equ	0x47		; 0=500kB/s, 1=250kB/s
CURGPL1	equ	0x48
CURGPL2	equ	0x49
CURDSP1	equ	0x4A
CURDSP2	equ	0x4B
CURDSPD	equ	0x4C
CURSPT	equ	0x4D
DSK0CHGFLG	equ	0x4E	; disk change flags for drives 0 and 1
DSK1CHGFLG	equ	0x4F


; --- Last variable space offset within bank 0 = 0x4F

; --- Following variables are located within shared space 0x70-0x7F
DELAY1		equ	0x70	; used within delay counting code
DELAY2		equ	0x71	; used within delay counting code
DELAY3		equ	0x72	; used within delay counting code



	org	0
	goto	main

	org	4
	goto	int_handler

Set144SpeedVals:
	movlw	GPL1_144
	movwf	CURGPL1
	movlw	GPL2_144
	movwf	CURGPL2
	movlw	DSPEC1_144
	movwf	CURDSP1
	movlw	DSPEC2_144
	movwf	CURDSP2
	movlw	DSPD_144
	movwf	CURDSPD
	movlw	SpT_144
	movwf	CURSPT
	return

Set720SpeedVals:
	movlw	GPL1_720
	movwf	CURGPL1
	movlw	GPL2_720
	movwf	CURGPL2
	movlw	DSPEC1_720
	movwf	CURDSP1
	movlw	DSPEC2_720
	movwf	CURDSP2
	movlw	DSPD_720
	movwf	CURDSPD
	movlw	SpT_720
	movwf	CURSPT
	return


main:
	movlw	1
	movwf	DSK0CHGFLG		; disk changed on startup
	movwf	DSK1CHGFLG

	clrf	DSK0SPD
	clrf	DSK1SPD		; setting speed 500kB/s

	; setting default speeds
	call	Set144SpeedVals

	clrf	DEBUGN

	; set timter TMR1
	BANKSEL	TMR1H
	clrf	TMR1H
	clrf	TMR1L
	BANKSEL	T1CON
	movlw	0x01
	movwf	T1CON
	BANKSEL	PIE1
	movlw	0x01
	movwf	PIE1
	BANKSEL	INTCON
	movlw	0xC0
	movwf	INTCON

	; increasing running speed to 32MHz (125 ns instruction cycle)
	BANKSEL	OSCCON
	movlw	0xf0			; 8Mhz and 32MHz option
	movwf	OSCCON			; PLL on, 32MHz

	BANKSEL	ANSELA
	clrf	ANSELA
	clrf	ANSELB
	clrf	ANSELD
	clrf	ANSELE		; all ports are digital outputs and inputs

	BANKSEL	TRISA
	movlw	0xFD		; RE1 as output, other pins as input
	movwf	TRISE
	movlw	0xFF
	movwf	TRISB
	movwf	TRISD		; ports A, B, D as input (W equals to 0xFF)
	clrf	TRISC		; port C as output
	clrf	TRISA		; port A as output

; DIAGNOSTIC PORT SETTING (port B as output)
clrf	TRISB

	BANKSEL	PORTA
	movwf	PORTA
	movwf	PORTB
	movwf	PORTC
	movwf	PORTD
	movwf	PORTE		; all ports A-D output high level (0xFF)

clrf	PORTB

	; clear controller status flag
	BANKSEL	PORTA			; select bank 0 where PORTx are located
	clrf	CTRSTAT
	clrf	CMDRUN
	clrf	MOTATO
	clrf	MOTBTO
	clrf	DRINUM
	clrf	CURDRV
	clrf	CURHOSTDRV

HostWaitMainLoop:
	bsf		INTCON,GIE	; disable interrupts
	movlw	0xAA
	movwf	COUVAL						; init interface counter value
	call	WaitForHostInitWrite			; wait for host write. Interrupts are enabled, thus host write may be missed by the MC.
											; To identify this condition CPU will need to read following status byte. If it's
											; 0xAA (or 0xAB, depending on request), then MC disables interrupts becoming fully available
											; for host requests and is ready to respond with valid data. If it's not 0xAA (0xAB), then
											; MC is in the progress of some command and if this command is not going
											; to be completed it should be terminated
											; WaitForHostInitWrite returns address bit in C, data written in W

	; process looks the following way:
	; 1. host issues write command with command needed to port base+0. If MC was in the process of some command, it aborts
	;	the command and returns to HostWaitMainLoop
	; 2. host writes parameters to the port base+1 of MC, each time receiving 0xAA+parameter #+1 (counter value). Host also can write 0xAA
	;	to port base+0 which returns this counter value, but does not request next parameter to be written or read.
	;	Host may reset command progress writing 0xAA to the base port+0. MC then returns 0xAA and goes to HostWaitMainLoop
	; 3. As soon as MC receives required set of parameters, it starts processing request. Host checks progress by writing to
	;	base port+1, if receives 0xFF this means that MC is not ready
	; 4. As soon as host reads next counter value, it performs data reads or writes as requested by the command. Each data read or
	; 	write will receive same counter number
	; 5. When required number of bytes were transferred, host starts to receive return information (first byte is next counter value,
	;	and second - return information if defined). If there're no parameters to return, host will not receive any valid data (most probably
	;	will receive 0xFF
	; 6. When all return data is read, MC returns to HostWaitMainLoop and waits for next command.
	; A. At any stage host may try to cancel current transaction (not MC command which runs if it was started) writing 0xAA to port base+0 in the loop
	;	and receiving 0xAA back if cancel is successful. Following, if hos twill need to cancel command running by MC, host issues INIHRD command,
	;	which resets MC and drives
	; B. If host writes to port base+0 value 0x80 when MC is waiting for command, host receives 0xAA back.


	; examination of the requested operation (WREG)
	; 0xAA = reset command progress and echo/looback (returns 0xAA to the host)
	; 0x00 = INIHRD, initialize FDC and FDDs (confirmation parameter 0xEC)
	; 0x01 = DSKIO read sector (drive#+head, cylinder, sector number; returns MC standard error data, see below)
	; 0x02 = DSKIO write sector (same as read sector)
	; 0x03 = DSKCHG check for disk change (drive#; returns 0xFF disk changed, 0x01 disk unchanged)
	; 0x04 = DSKFMT disk format (confirmation parameter 0xEC, and then drive#)
	; 0x05 = DRIVES return number of physical drives and currently selected drive (0-3)
	; > 0x05 = reset request, return 0x55 to the host and restart sensing new request

	btfsc	STATUS,C					; A0 address line state during host request
	goto	HostWaitMainLoop			; access to port base+1 is not assumed this time

	bcf		INTCON,GIE	; disable interrupts

	movwf	HOSTCMD
	sublw	0xAA
	btfsc	STATUS,Z
	goto	NoCommandToRun

	movf	HOSTCMD,W
	btfsc	STATUS,Z
	goto	INITHRD
	decf	WREG,1
	btfsc	STATUS,Z
	goto	DSKIO
	decf	WREG,1
	btfsc	STATUS,Z
	goto	DSKIO
	decf	WREG,1
	btfsc	STATUS,Z
	goto	DSKCHG
	decf	WREG,1
	btfsc	STATUS,Z
	goto	DSKFMT
	decf	WREG,1
	btfsc	STATUS,Z
	goto	DRIVES

NoCommandToRun:
	movlw	0xAA
	movwf	COUVAL
	call	ReturnByteToHost			; returns 0xAA to the host = "loopback", or "no command to run"
	goto	HostWaitMainLoop

; --- INIHRD host routine
INITHRD:
	movlw	0xAB						; command accepted: now stage 2 to get first parameter from host
	movwf	COUVAL
	call	RetStatToHostAndWaitForHostWrite	; return 0xAB to the host
												; and get next byte from the host
	btfss	STATUS,C					; address A0 is set to 1 (reset MC operation process)
	goto	NoCommandToRun				; in this case return 0xAA

	sublw	0xEC						; pass word to start initialization of the FDD hardware
	btfss	STATUS,Z
	goto	NoCommandToRun

	call	InitHardware

	movlw	0xAC
	movwf	COUVAL
	movlw	0x80
	call	RetStatToHostAndWaitForHostRead			; return 0x80 to the host
	goto	HostWaitMainLoop

; --- DRIVES host routine
DRIVES:
	movlw	0xAB						; command accepted
	movwf	COUVAL
	movf	DRINUM,W

	call	RetStatToHostAndWaitForHostRead	; return status (0xAB) and then return DRINUM
	movlw	0xAC						; next return byte
	movwf	COUVAL
	movf	CURHOSTDRV,W
	call	RetStatToHostAndWaitForHostRead	; return status (0xAC) and then return DRINUM
	goto	HostWaitMainLoop

; --- DSKIO host routine
DSKIO:
	movlw	0xAB						; command accepted: now stage 2 to get first parameter from host
	movwf	COUVAL
	call	RetStatToHostAndWaitForHostWrite	; return 0xAB to the host - operation is in progress now
									; and get first parameter: drive # (0=A:) and head (in FDC format)
	btfss	STATUS,C				; A0 address line state during host request
	goto	NoCommandToRun			; access to port base+1 is not assumed this time, aborting MC command
	movwf	DRVPRM

	movlw	0xAC						; command accepted: now stage 3 to get second parameter from host
	movwf	COUVAL
	call	RetStatToHostAndWaitForHostWrite	; return 0xAC to the host
									; and get second parameter: cylinder#
	btfss	STATUS,C				; A0 address line state during host request
	goto	NoCommandToRun			; access to port base+1 is not assumed this time, aborting MC command
	movwf	CYLPRM

	movlw	0xAD						; command accepted: now stage 4 to get third parameter from host
	movwf	COUVAL
	call	RetStatToHostAndWaitForHostWrite	; return 0xAD to the host
									; and get third parameter: sector# (track starts with sector 0x01)
	btfss	STATUS,C				; A0 address line state during host request
	goto	NoCommandToRun			; access to port base+1 is not assumed this time, aborting MC command
	movwf	SECPRM

	; after successfully getting third parameter MC command starts to execute. Host should monitor
	; return at the base port+1 to become 0xAE for data transfer or for 0xAF for status transfer

	; turn motor on
	movf	DRVPRM,W
	andlw	0x01
;	movwf	CURDRV					; drive number (0/1)
	call	CheckSingleDrive
	call	MotorOnSelect

	movlw	2
	movwf	SPDCOU					; trying two speeds

SpeedTryLoop:
	; take remembered value for speed for specific drive (0/1)
	movf	DSK0SPD,W
	btfsc	CURDRV,0
	movf	DSK1SPD,W

	btfss	WREG,0
	goto	Set144
	call	Set720SpeedVals
	goto	SpeedsSet
Set144:
	call	Set144SpeedVals

SpeedsSet:
	; set speeds for specific drive according to its remembered mode
	call	SetSpeeds

	movf	HOSTCMD,W
	decf	WREG,W
	btfsc	STATUS,Z
	goto	DoNotCheckWriteProtected

	; write operation - check for disk write protection issuing SENSE DRIVE STATUS command
	call	NewFDCRequest			; clear MC FDC queue
	movlw	0x04
	call	AddToFDCQueue
	movf	DRVPRM,W
	call	AddToFDCQueue
	call	RunCommand

	movf	FDCERR,1
	btfss	STATUS,Z
	goto	DSKIORetUnkErr

NoSenseDSError:
	movf	0x60,W					; get value of ST3
	andlw	0x40
	btfsc	STATUS,Z
	goto	DoNotCheckWriteProtected

	; diskette is write protected
	clrw							; WRITE PROTECT error
	goto	DSKIORetStat

DoNotCheckWriteProtected:
	; check if disk was changed
	; if disk is not changed, then it does not matter if next seek command will really seek and set disk change bit
	; but if it's changed, then we seek to cylinder 1, then perform recalibrate, and check "disk change" bit again
	; if it's still set, it means there's no disk in drive - and we return with error
	call	ReadDiskChange
	btfss	TEMPREQ,7
	goto	NotChanged			; disk change bit set - proceed with seek and see if it will be still set

	movf	CYLPRM,W
	movwf	FMTPR1				; save current cylinder number

	movlw	0x01
	movwf	CYLPRM
	call	SeekCommand				; Cylinder 1

	call	Recalibrate				; recalibrate
	btfsc	FDCERR,0
	goto	DSKIORetUnkErr			; recalibration error

	movf	FMTPR1,W
	movwf	CYLPRM

	call	ReadDiskChange
	btfsc	TEMPREQ,7
	goto	DSKIORetUnkErr			; disk change bit set after seek - no diskette present - return

NotChanged:
	; proceed with seek and data read/write
	call	SeekCommand				; nothing happens if heads are already positioned there
	movf	FDCERR,1
	btfsc	STATUS,Z
	goto	NoSeekError
DSKIORetSeekErr:
	movlw	0x06					; SEEK ERROR
	goto	DSKIORetStat

NoSeekError:
	; and try reading data
	call	NewFDCRequest			; clear MC FDC queue
	movf	HOSTCMD,W
	decf	WREG,W
	movlw	0x46					; HOSTSCMD=1 -- read operation 0x46
	btfss	STATUS,Z
	movlw	0x45					; HOSTSCMD=2 -- write operation 0x45
	call	AddToFDCQueue
	movf	DRVPRM,W
	call	AddToFDCQueue
	movf	CYLPRM,W
	call	AddToFDCQueue
	movf	DRVPRM,W
	asrf	WREG,1
	asrf	WREG,1
	call	AddToFDCQueue
	movf	SECPRM,W
	call	AddToFDCQueue
	movlw	0x02
	call	AddToFDCQueue
	movf	SECPRM,W				; last sector# on the track to transfer (the same as sector# to read)
	call	AddToFDCQueue
	movlw	CURGPL1					; GAP1 default size
	call	AddToFDCQueue
	movlw	0xFF					; Data length (default 0xFF)
	call	AddToFDCQueue

	movlw	0xAE
	movwf	COUVAL					; data transfer stage (0xAE)

	movf	HOSTCMD,W
	decf	WREG,W
	btfsc	STATUS,Z
	goto	DoNotLoadBufferFromHost	

	call	LoadDataToBufferFromHost		; host will write data tp MC (FDD)

	btfss	STATUS,C				; A0 address line state during host request
	goto	NextSpeed				; try another speed setting

DoNotLoadBufferFromHost:
	call	RunCommandWithData
	movf	FDCERR,1
	btfsc	STATUS,Z
	goto	NoAccessError

NextSpeed:
	; we are here because of error occured
	; change the speed, and try again
	decfsz	SPDCOU,1
	goto	ChangeSpeed
	goto	DSKIORetUnkErr

ChangeSpeed:
	movlw	0x01
	btfss	CURDRV,0
	goto	ChgDrv0
	; drive 1
	xorwf	DSK1SPD,1
	goto	SpeedTryLoop
ChgDrv0:
	xorwf	DSK0SPD,1
	goto	SpeedTryLoop

	; both speed attempts failed
DSKIORetUnkErr:
	movlw	0x0C					; "Unknown error"
	goto	DSKIORetStat

NoAccessError:
	call	DecodeError				; returns 0xCC if no error
	btfss	WREG,7					; bit 7 = 1 - no error occured
	goto	NextSpeed				; try another speed setting

	btfss	HOSTCMD,0				; bit 0 of command = 0 -> write op (10b) (do not wait host for reading data)
	goto	DSKIORetStat

	call	TransferDataToHostFromBuffer	; host will read data from MC (FDD)

	btfss	STATUS,C				; A0 address line state during host request
	goto	NextSpeed				; try another speed setting

DSKIOSuccRetStat:
	; save speed ID of successful r/w operation
	movf	SPDCOU,W
	btfss	CURDRV,0
	goto	SavDrv0
	; drive 1
	movwf	DSK1SPD
	goto	DSKIOExitSuccess
SavDrv0:
	movwf	DSK0SPD
DSKIOExitSuccess:
	movlw	0xCC

DSKIORetStat:
	; here host is trying to read return byte and then will read status byte

; DIAGNOSTIC DISPLAY - DSKIO return code
;movwf	PORTB

	movwf	SECPRM					; preserve return byte
	call	MotorOff

	movlw	0xAF
	movwf	COUVAL
	movf	SECPRM,W				; restore return byte
	call	RetStatToHostAndWaitForHostRead
	goto	HostWaitMainLoop

; Routine seeks to specified cylinder and remembers 
; in: (DRVPRM) = drive number in FDC format, (CYLPRM) = cylinder to seek to
; updates HD*POS value to identify current head position of the drive
SeekCommand:
	call	ReadDiskChange
	btfsc	TEMPREQ,7
	call	DiskChangeFlagSet		; set disk change flag for specified drive for DSKCHG sub

	call	NewFDCRequest			; clear MC FDC queue
	movlw	0x0F
	call	AddToFDCQueue
	movf	DRVPRM,W
	call	AddToFDCQueue
	movf	CYLPRM,W
	call	AddToFDCQueue
	call	RunCommand
	call	WaitSeekComplete
	return

; in: CURDRV = current drive ID (bit 0)
; out: none
DiskChangeFlagSet:
	btfss	CURDRV,0
	goto	FlagSetDrv0

	; drive 1 disk change flag setting
	bsf		DSK1CHGFLG,0
	return
FlagSetDrv0:
	bsf		DSK0CHGFLG,0
	return

; in: CURDRV = current drive ID (bit 0)
; out: none
DiskChangeFlagReset:
	btfss	CURDRV,0
	goto	FlagResetDrv0

	; drive 1 disk change flag setting
	clrf	DSK1CHGFLG
	return
FlagResetDrv0:
	clrf	DSK0CHGFLG
	return

; in: CURDRV = current drive ID (bit 0)
; out: W = flag (bit 0)
DiskChangeFlagGet:
	btfss	CURDRV,0
	goto	FlagGetDrv0

	; drive 1 disk change flag setting
	movf	DSK1CHGFLG,W
	return
FlagGetDrv0:
	movf	DSK0CHGFLG,W
	return


; -------------- Read sector data from host and write to MC buffer ------------------
LoadDataToBufferFromHost:
	; initialize variables for data transfer
	clrf	RECLENL
	movlw	0x02
	movwf	RECLENH			; byte count in the record (512)

	movlw	0x01			; bank 4
	movwf	FSR0H	
	movlw	0x20			; start buffer pointer
	movwf	FSR0L

DIOWdata:
	call	RetStatToHostAndWaitForHostWrite	; returns 0xAE to host and waits for host write
	btfss	STATUS,C				; A0 address line state during host request
	return							; access to port base+1 is not assumed this time, aborting MC command

	movwi	FSR0++						; write data to MC buffer

	decfsz	RECLENL,1
	goto	NextPtrW
	decf	RECLENH,1		; increase byte count
	bsf		STATUS,C		; set C flag (like access to base port+1)
	btfsc	STATUS,Z
	return

NextPtrW:
	movf	FSR0,W
	andlw	0x7F			; mask all bits except bit 7
	sublw	0x60			; end of buffer chunk = 0x5F
	btfss	STATUS,Z
	goto	DIOWdata		; still within current buffer chunk

	movlw	0x40
	addwf	FSR0L,1
	movlw	0x00
	addwfc	FSR0H,1			; increment 16-bit pointer by 0x40 (start of next chunk)
	goto	DIOWdata
; ************** Read sector data from host and write to MC buffer ******************

; -------------- Read sector data from MC buffer and send it to host ------------------
TransferDataToHostFromBuffer:
	; initialize variables for data transfer
	clrf	RECLENL
	movlw	0x02
	movwf	RECLENH			; byte count in the record (512)

	movlw	0x01			; bank 4
	movwf	FSR0H	
	movlw	0x20			; start buffer pointer
	movwf	FSR0L

DIORdata:
	moviw	FSR0++			; read data from buffer and increase FSR0
	call	RetStatToHostAndWaitForHostRead		; returns 0xAE to host and waits for host read
	btfss	STATUS,C				; A0 address line state during host request
	return							; access to port base+1 is not assumed this time, aborting MC command

	decfsz	RECLENL,1
	goto	NextPtrR
	decf	RECLENH,1		; decrease byte count
	bsf		STATUS,C		; set C flag (like access to base port+1)
	btfsc	STATUS,Z
	return

NextPtrR:
	movf	FSR0,W
	andlw	0x7F			; mask all bits except bit 7
	sublw	0x60			; end of buffer chunk = 0x5F
	btfss	STATUS,Z
	goto	DIORdata		; still within current buffer chunk

	movlw	0x40
	addwf	FSR0L,1
	movlw	0x00
	addwfc	FSR0H,1			; increment 16-bit pointer by 0x40 (start of next chunk)
	goto	DIORdata
; ************** Read sector data from MC buffer and send it to host ******************

DecodeError:
	; Errors returned:
	; [0 = write protected] - not returned here, needs sense status command
	; [2 = not ready] - not returned here
	; 4 = crc error
	; 6 = seek error
	; 8 = record not found
	; 10 = write fault
	; 12 = other error
	; 0xCC = no error

	movf	FDCERR,1
	btfss	STATUS,Z
	goto	UnkError				; most probably timeout occured, sense data returned by FDC may not be valid

	; ST0
	btfsc	0x60,4					; first byte of the result data, ST0, bit 4 = "CHECK EQUIPMENT"
	goto	SeekError

	; ST1
	btfsc	0x61,5					; second byte of the result data, ST1, bit 5 = "DATA ERROR"
	goto	CrcError
	btfsc	0x61,2					; second byte of the result data, ST1, bit 2 = "NO DATA"
	goto	RecordNotFound
	btfsc	0x61,4					; second byte of the result data, ST1, bit 4 = "OVERRUN"
	goto	UnkError
	btfsc	0x61,1					; second byte of the result data, ST1, bit 1 = "NOT WRITABLE"
	goto	WriteFault
	btfsc	0x61,0					; second byte of the result data, ST1, bit 0 = "MISSING ADDRESS MARK"
	goto	RecordNotFound
;	btfsc	0x61,7					; second byte of the result data, ST1, bit 7 = "END OF CYLINDER"
;	goto	RecordNotFound

	; ST2
	btfsc	0x62,5					; second byte of the result data, ST1, bit 5 = "DATA ERROR"
	goto	CrcError
	btfsc	0x62,6					; second byte of the result data, ST1, bit 6 = "CONTROL MARK"
	goto	RecordNotFound
	btfsc	0x62,4					; second byte of the result data, ST1, bit 6 = "WRONG CYLINDER"
	goto	SeekError
	btfsc	0x62,1					; second byte of the result data, ST1, bit 6 = "BAD CYLINDER"
	goto	SeekError
	btfsc	0x62,4					; second byte of the result data, ST1, bit 6 = "NISSING ADDRESS MARK"
	goto	RecordNotFound

	movlw	0xCC					; no error
	return
UnkError:
	movlw	0x0C
	return
SeekError:
	movlw	0x06
	return
RecordNotFound:
	movlw	0x08
	return
CrcError:
	movlw	0x04
	return
WriteFault:
	movlw	0x0A
	return


DSKCHG:
	movlw	0xAB						; command accepted: now stage 2 to get first parameter from host
	movwf	COUVAL
	call	RetStatToHostAndWaitForHostWrite	; return 0xAB to the host - operation is in progress now
												; and get first parameter: drive # (0=A:)
	btfss	STATUS,C				; A0 address line state during host request
	goto	NoCommandToRun			; access to port base+1 is not assumed this time, aborting MC command

	andlw	0x01					; drive number
;	movwf	CURDRV
	call	CheckSingleDrive
	movwf	DRVPRM
	call	MotorOnSelect
	call	ReadDiskChange			; check FDD's disk change bit
	call	MotorOff

	movlw	0xAF					; data return stage (0xAF)
	movwf	COUVAL

	btfsc	TEMPREQ,7
	goto	DSKChanged

	call	DiskChangeFlagGet
	btfsc	WREG,0
	goto	DSKChanged

	movlw	0x01					; disk unchanged
	goto	RetChg
DSKChanged:
	call	DiskChangeFlagReset
	; now clear the disk change bit by seeking to cylinder 10 first, and the cylinder 0
;	call	SoftReset
	call	MotorOnSelect
	movlw	0x01
	movwf	CYLPRM
	call	SeekCommand				; Cylinder 1
	clrf	CYLPRM
	call	SeekCommand				; Cylinder 0
	call	MotorOff
	movlw	0xFF					; disk changed

RetChg:
	call	RetStatToHostAndWaitForHostRead		; write statsu byte and then disk change code to the host
	goto	HostWaitMainLoop

; ---------------------------------------
; reads disk change attribute from LDCR - currently selected drive
; returns: TEMPREQ=LDCR byte read
ReadDiskChange:
	; read LDCR, bit 7
	BANKSEL	TRISA
	movlw	0xFF
	movwf	TRISA					; put port A into input mode
	BANKSEL	PORTA

	clrf	PORTE					; activate /LDCR
	movlw	0xF0					; /LDOR=1, /RD=0, A=0, /CS=1: read FDC control register
	movwf	PORTC
	nop
	movf	PORTA,W
	movwf	TEMPREQ
	movlw	0xF8
	movwf	PORTC					; release /RD
	BANKSEL	TRISA
	clrf	TRISA					; set port A back to output mode
	BANKSEL	PORTA
	movlw	0xFF
	movwf	PORTE					; release /LDCR

	; set disk change flag if diskchange bit is set
	btfsc	TEMPREQ,7
	call	DiskChangeFlagSet		; set disk change flag for DSKCHG sub
	return
; ---------------------------------------

DSKFMT:
	movlw	0xAB					; command accepted: now stage 2 to get first parameter from host
	movwf	COUVAL
	call	RetStatToHostAndWaitForHostWrite	; return 0xAB to the host - operation is in progress now
												; and get next byte write from the host
	btfss	STATUS,C				; address A0 is set to 1 (reset MC operation process)
	goto	NoCommandToRun			; in this case return 0xAA

	movwf	DRVPRM
	andlw	0x01
;	movwf	CURDRV					; current drive
	call	CheckSingleDrive

	movlw	0xAC					; now stage 3 to get second parameter - confirmation code
	movwf	COUVAL

	call	RetStatToHostAndWaitForHostWrite	; return 0xAC to the host - operation is in progress now
											; and get second: format parameter
	btfss	STATUS,C				; A0 address line state during host request
	goto	NoCommandToRun			; access to port base+1 is not assumed this time, aborting MC command

	movwf	FMTPRM					; format parameter (1-standard 1.44, 2-compatible 1.44, , 3-compatible 720)

	movlw	0xAD					; now stage 4 to get third parameter - confirmation codes
	movwf	COUVAL

	call	RetStatToHostAndWaitForHostWrite	; return 0xAC to the host - operation is in progress now
											; and get second: confirmation code
	btfss	STATUS,C				; A0 address line state during host request
	goto	NoCommandToRun			; access to port base+1 is not assumed this time, aborting MC command

	sublw	0xEC					; pass word
	btfss	STATUS,Z
	goto	NoCommandToRun

	movlw	0xAE
	movwf	COUVAL
	movlw	0x80
	call	RetStatToHostAndWaitForHostRead		; return 0x80 to the host - confirmation that format has started

	; set drive speed
	call	Set144SpeedVals
	movf	FMTPRM,W
	sublw	0x03					; compatible 720kB
	btfsc	STATUS,Z
	call	Set720SpeedVals

	call	InitDrive				; initialize drive (CURDRV) to format disk in, and sets speeds

	clrf	CYLPRM					; init cylinder#=0
	clrf	SECPRM					; init head#=0

	; checking for diskette write protection - issuing SENSE DRIVE STATUS command
	call	NewFDCRequest			; clear MC FDC queue
	movlw	0x04
	call	AddToFDCQueue
	movf	DRVPRM,W
	call	AddToFDCQueue
	call	RunCommand

	movf	FDCERR,1
	btfss	STATUS,Z
	goto	DSKIORetUnkErr			; "Unknown error"

	movf	0x60,W					; get value of ST3
	andlw	0x40
	btfsc	STATUS,Z
	goto	FormatLoop

	; diskette is write protected
	clrw							; WRITE PROTECT error
	goto	DSKIORetStat

FormatLoop:
	; proceed with seek to the specified cylinder
	call	SeekCommand
	movf	FDCERR,1
	btfss	STATUS,Z
	goto	DSKIORetSeekErr			; SEEK ERROR

FormatSeekOk:
	call	NewFDCRequest
	movlw	0x4D					; FORMAT TRACK command, MFM mode
	call	AddToFDCQueue
	movf	CURDRV,W
	iorwf	SECPRM,W				; OR with head number (in bit 2)
	call	AddToFDCQueue
	movlw	0x02					; Sector size = 512 bytes
	call	AddToFDCQueue
	movf	CURSPT,W				; Sectors per track (current value)
	call	AddToFDCQueue
	movlw	CURGPL2					; gap length 2 for formatting
	call	AddToFDCQueue
	movlw	0xE5					; Filler byte for data
	call	AddToFDCQueue

	movf	CYLPRM,W
	call	FillBufForFormat		; build tables to format specified track, should be run
									; AFTER format command buffer is filled with data (uses byte at offset 0x51)
									; it fills data for 18 consequtive sectors, even if we format 9 SpT

	call	RunCommandWithData
	movf	FDCERR,1
	btfsc	STATUS,Z
	goto	TrackFormatOk
FormatRetWriteFaultErr:
	movlw	0x0A					; WRITE FAULT
	goto	DSKIORetStat

TrackFormatOk:
	; next head
	movf	SECPRM,1
	btfss	STATUS,Z
	goto	NextCyl

	movlw	0x04
	addwf	SECPRM,1				; adds to the second FORMAT command parameter (bit 2++)
	goto	FormatSeekOk			; same track, side 1

NextCyl:
	clrf	SECPRM
	incf	CYLPRM,1
	movf	CYLPRM,W
	sublw	0x50					; track 79 is the last one
	btfss	STATUS,Z
	goto	FormatLoop

; --- Now writing control blocks: Boot sector and FAT
	; format options: 1=standard 1.44, 2=compatible 1.44, 3=compatible 720
	movf	FMTPRM,W				; starts with 1 (not 0)
	decfsz	WREG,1
	goto	Compat144_720Format

	; standard 1.44 format (1 sectors per cluster, 9 FAT sectors, GR8BIT stack at 0C382h)
	movlw	0x21
	movwf	FMTPR1					; control block sector count (33)
	movlw	0x20					; first FAT is at sector 1
	movwf	FMTPR2					; first FAT counter
	movlw	0x17					; seconf FAT is at sector 10 (=1+9)
	movwf	FMTPR3					; second FAT counter
	goto	ProceedWrite

Compat144_720Format:
	; we are here for options 2 and 3, former is compatible 1.44 and
	; latter is compatible 720, which have same control structures' allocation
	; boot=0, fat1=1, fat2=4, dir=7, data=14
	; in both cases (SpT=18 and SpT=9) structure sare located within cylinder 0

	; standard 1.44 format (4 sectors per cluster, 3 FAT sectors, 720K compatible stack location)
	movlw	0x0E
	movwf	FMTPR1					; control block sector count (33)
	movlw	0x0D					; first FAT is at sector 1
	movwf	FMTPR2					; first FAT counter
	movlw	0x0A					; second FAT is at sector 4 (=1+3)
	movwf	FMTPR3					; second FAT counter
	goto	ProceedWrite

ProceedWrite:
	movf	FMTPR1,W				; sector count (0x21 for standard and 0x0E for compatible)
	movwf	FMTTMP1					; current logical sector number in countdown mode
	clrf	CYLPRM					; current cylinder#
	movlw	0x01
	movwf	SECPRM					; currect sector number
	movf	CURDRV,W
	movwf	DRVPRM					; head bit in CURDRV = 0 (head 0)

WriteBootFat:
	; proceed with seek to the specified cylinder
	call	SeekCommand
	movf	FDCERR,1
	btfss	STATUS,Z
	goto	DSKIORetSeekErr			; SEEK ERROR

	call	ClearBufForFormat

	movf	FMTPR1,W				; sector count (0x21 for standard and 0x0E for compatible)
	subwf	FMTTMP1,W
	btfss	STATUS,Z
	goto	NotBootSector

	movlw	0x01
	movwf	FSR0H
	movlw	0x20
	movwf	FSR0L				; set up indirect pointer to the buffer

	movlw	0xb7
	movwf	FMTTMP2				; bytes in boot sector data

BootFill:
	movf	FMTTMP2,W
	sublw	0xb8
	movwf	FMTTMP3

	movlw	LOW BootSectTable
	addwf	FMTTMP3,1
	movlw	HIGH BootSectTable
	btfsc	STATUS,C
	addlw	1
	movwf	PCLATH
	movf	FMTTMP3,W

	call	BootSectTable
	movwi	FSR0++

	decfsz	FMTTMP2,1
	goto	NextBootPtr

	; now fille boot sector with proper disk configuration data
	movf	FMTPRM,W
	decfsz	WREG,1
	goto	CompatConfig

	movlb	2
	movlw	0x01		; Standard config: 1 sector per cluster, 224 dir entries, 9 sectors per FAT
	movwf	0x2D		; cluster size
	movlw	0xE0
	movwf	0x31		; dir entries
	movlw	0x09
	movwf	0x36		; size of FAT
	movlb	0
	goto	NotFAT

CompatConfig:
	decfsz	WREG,1
	goto	Compat720

	movlb	2
	movlw	0x04		; Compatible config 1.44M: 4 sectors per cluster, 112 dir entries, 3 sectors per FAT
	movwf	0x2D		; cluster size
	movlw	0x70
	movwf	0x31		; dir entries
	movlw	0x03
	movwf	0x36		; size of FAT
	movlb	0
	goto	NotFAT

Compat720:
	movlb	2
	movlw	0x02		; Compatible config 720K: 2 sectors per cluster, 112 dir entries, 3 sectors per FAT
	movwf	0x2D		; cluster size
	movlw	0x70
	movwf	0x31		; dir entries
	movlw	0x03
	movwf	0x36		; size of FAT
	movlw	0xA0
	movwf	0x33
	movlw	0x05
	movwf	0x34		; sectors on disk (0x5A0)
	movlw	0xF9
	movwf	0x35		; 720K media descriptor (0xF9)
	movlw	0x09
	movwf	0x38		; number of sectors per track (0x9)
	movlb	0
	goto	NotFAT

NextBootPtr:
	movf	FSR0,W
	andlw	0x7F			; mask all bits except bit 7
	sublw	0x60			; end of buffer chunk = 0x5F
	btfss	STATUS,Z
	goto	BootFill	; still within current buffer chunk

	movlw	0x40
	addwf	FSR0L,1
	movlw	0x00
	addwfc	FSR0H,1			; increment 16-bit pointer by 0x40 (start of next chunk)
	goto	BootFill

NotBootSector:
	movf	FMTPR2,W				; first FAT pointer count (0x20 for standard and 0x0D for compatible)
	subwf	FMTTMP1,W
	btfsc	STATUS,Z
	goto	FATSector
	movf	FMTPR3,W				; second FAT pointer count (0x17 for standard and 0x0A for compatible)
	subwf	FMTTMP1,W
	btfss	STATUS,Z
	goto	NotFAT

FATSector:
	; select propert media descriptor for FAT copies
	movf	FMTPRM,W				; starts with 1 (not 0)
	decfsz	WREG,1
	goto	DescF9
	movlw	0xF0
	goto	EndDesc
DescF9:
	movlw	0xF9
EndDesc:
;	movlw	0xF0			; FAT descriptor
	movlb	2
	movwf	0x20
	movlw	0xFF
	movwf	0x21
	movwf	0x22
	movlb	0

NotFAT:
	call	NewFDCRequest			; clear MC FDC queue
	movlw	0x45
	call	AddToFDCQueue			; FDC write command
	movf	DRVPRM,W
	call	AddToFDCQueue
	movf	CYLPRM,W
	call	AddToFDCQueue
	movf	DRVPRM,W
	asrf	WREG,1
	asrf	WREG,1
	call	AddToFDCQueue
	movf	SECPRM,W
	call	AddToFDCQueue
	movlw	0x02
	call	AddToFDCQueue
	movf	SECPRM,W				; last sector# on the track to transfer (the same as sector# to read)
	call	AddToFDCQueue
	movlw	CURGPL1					; GAP1 default size
	call	AddToFDCQueue
	movlw	0xFF					; Data length (default 0xFF)
	call	AddToFDCQueue
	call	RunCommandWithData

	movf	FDCERR,1
	btfss	STATUS,Z
	goto	FormatRetWriteFaultErr	; "WRITE FAULT"

	incf	SECPRM,1

	; setting max sector number +1
	movf	FMTPRM,W				; starts with 1 (not 0)
	decfsz	WREG,1
	goto	CompatSectors
	movlw	0x13					; 18 sectors per track
	goto	EndSectors
CompatSectors:
	movlw	0x0A					; 9 sectors per track
EndSectors:
;	movlw	0x13					; 18 sectors (+1 = overflow)
	subwf	SECPRM,W
	btfss	STATUS,Z
	goto	NoNewHead

	movlw	0x01
	movwf	SECPRM

	; next head here
	btfss	DRVPRM,2
	goto	NextHeadSameTrack

	; next track here
	bcf		DRVPRM,2
	incf	CYLPRM,1
	goto	NoNewHead

NextHeadSameTrack:
	bsf		DRVPRM,2

NoNewHead:
	decfsz	FMTTMP1,1
	goto	WriteBootFat

	; format ended successfully
	goto	DSKIOSuccRetStat

; ------------------- Boot sector data (called with index in W) --------------------
BootSectTable:			; disk configuration data and boot loader
	movwf	PCL
	; lookup table
	retlw	0xeb
	retlw	0xfe
	retlw	0x00
	retlw	0x47		; "GR8BIT10" format identifier
	retlw	0x52
	retlw	0x38
	retlw	0x42
	retlw	0x49
	retlw	0x54
	retlw	0x31
	retlw	0x30
	retlw	0x00		; bytes per sector (2)
	retlw	0x02
FormatChg1:	; offset 0x2D
	retlw	0x00		; cluster size in sectors (1) - 0x01 for standard and 0x04 for compatible formats
	retlw	0x01		; number of unused sectors (2)
	retlw	0x00
	retlw	0x02		; number of FATs (1)
FormatChg2:	; offset 0x31
	retlw	0x00		; number of directory entries (2) - 0xE0 for standard and 0x70 for compatible formats
	retlw	0x00
FormatChg3:	; offset 0x33
	retlw	0x40		; number of sectors per disk (2)
	retlw	0x0b
FormatChg4:	; offset 0x35
	retlw	0xf0		; media ID (1)
FormatChg5:	; offset 0x36
	retlw	0x00		; size of FAT in sectors (2) - 0x09 for standard and 0x03 for compatible formats
	retlw	0x00
FormatChg6:	; offset 0x38
	retlw	0x12		; number of sectors per track (2)
	retlw	0x00
	retlw	0x02		; number of sides (2)
	retlw	0x00
	retlw	0x00		; number of hidden sectors (2)
	retlw	0x00
	retlw	0xd0
	retlw	0xed
	retlw	0x53
	retlw	0x59
	retlw	0xc0
	retlw	0x32
	retlw	0xd0
	retlw	0xc0
	retlw	0x36
	retlw	0x56
	retlw	0x23
	retlw	0x36
	retlw	0xc0
	retlw	0x31
	retlw	0x1f
	retlw	0xf5
	retlw	0x11
	retlw	0xab
	retlw	0xc0
	retlw	0x0e
	retlw	0x0f
	retlw	0xcd
	retlw	0x7d
	retlw	0xf3
	retlw	0x3c
	retlw	0xca
	retlw	0x63
	retlw	0xc0
	retlw	0x11
	retlw	0x00
	retlw	0x01
	retlw	0x0e
	retlw	0x1a
	retlw	0xcd
	retlw	0x7d
	retlw	0xf3
	retlw	0x21
	retlw	0x01
	retlw	0x00
	retlw	0x22
	retlw	0xb9
	retlw	0xc0
	retlw	0x21
	retlw	0x00
	retlw	0x3f
	retlw	0x11
	retlw	0xab
	retlw	0xc0
	retlw	0x0e
	retlw	0x27
	retlw	0xcd
	retlw	0x7d
	retlw	0xf3
	retlw	0xc3
	retlw	0x00
	retlw	0x01
	retlw	0x58
	retlw	0xc0
	retlw	0xcd
	retlw	0x00
	retlw	0x00
	retlw	0x79
	retlw	0xe6
	retlw	0xfe
	retlw	0xfe
	retlw	0x02
	retlw	0xc2
	retlw	0x6a
	retlw	0xc0
	retlw	0x3a
	retlw	0xd0
	retlw	0xc0
	retlw	0xa7
	retlw	0xca
	retlw	0x22
	retlw	0x40
	retlw	0x11
	retlw	0x85
	retlw	0xc0
	retlw	0xcd
	retlw	0x77
	retlw	0xc0
	retlw	0x0e
	retlw	0x07
	retlw	0xcd
	retlw	0x7d
	retlw	0xf3
	retlw	0x18
	retlw	0xb4
	retlw	0x1a
	retlw	0xb7
	retlw	0xc8
	retlw	0xd5
	retlw	0x5f
	retlw	0x0e
	retlw	0x06
	retlw	0xcd
	retlw	0x7d
	retlw	0xf3
	retlw	0xd1
	retlw	0x13
	retlw	0x18
	retlw	0xf2
	retlw	0x42
	retlw	0x6f
	retlw	0x6f
	retlw	0x74
	retlw	0x20
	retlw	0x65
	retlw	0x72
	retlw	0x72
	retlw	0x6f
	retlw	0x72
	retlw	0x0d
	retlw	0x0a
	retlw	0x50
	retlw	0x72
	retlw	0x65
	retlw	0x73
	retlw	0x73
	retlw	0x20
	retlw	0x61
	retlw	0x6e
	retlw	0x79
	retlw	0x20
	retlw	0x6b
	retlw	0x65
	retlw	0x79
	retlw	0x20
	retlw	0x66
	retlw	0x6f
	retlw	0x72
	retlw	0x20
	retlw	0x72
	retlw	0x65
	retlw	0x74
	retlw	0x72
	retlw	0x79
	retlw	0x0d
	retlw	0x0a
	retlw	0x00
	retlw	0x00
	retlw	0x4d
	retlw	0x53
	retlw	0x58
	retlw	0x44
	retlw	0x4f
	retlw	0x53
	retlw	0x20
	retlw	0x20
	retlw	0x53
	retlw	0x59
	retlw	0x53
; ******************* Boot sector data (called with index in W) ********************

; ------------------- Initialize hardware (called from INIHRD) ---------------------
InitHardware:
	; --- Wait until system reset condition is cleared
LoopReset:
	movf	PORTE,W
	andlw	0xC			; RE3/RE2
	btfss	STATUS,Z
	goto	LoopReset	; loop until both are 0 (inactive)

	movlw	0xFF
	call	Wait_ms		; wait 255 ms

	call	SoftReset

	clrf	DRINUM
	clrf	CURDRV
	call	InitDrive	; init drive 0 (A)
	btfss	STATUS,C
	incf	DRINUM,1	; now drive count = 1
	incf	CURDRV,1
	call	InitDrive	; init drive 1 (B)
	btfss	STATUS,C
	incf	DRINUM,1	; now drive count = 2

	clrf	CURDRV		; by default after hardware init drive 0 is selected
	clrf	CURHOSTDRV
	return
; ******************* Initialize hardware (called from INIHRD) *********************

SetSpeeds:
	; set channel speed
	movf	CURDSPD,W
	movwf	PORTA
	clrf	PORTE		; activate /LDCR

	movlw	0xE8		; /LDOR=1, /WR=0, A=0, /CS=1: write FDD controller operations register, base mode
	movwf	PORTC
	nop
	movlw	0xF8
	movwf	PORTC		; release /WR
	movlw	0xFF
	movwf	PORTE		; release /LDCR
	movwf	PORTA		; release data bus

; --- Specify (global attribute about disk drives' modes and times)
	call	NewFDCRequest
	movlw	0x03		; SPECIFY command
	call	AddToFDCQueue
	movf	CURDSP1,W
	call	AddToFDCQueue
	movf	CURDSP2,W
	call	AddToFDCQueue
	call	RunCommand
	return

; SoftReset:

; --- SOFT RESET OF FDD CONTROLLER ---
SoftReset:
	; Soft resetting FDD controller: LDOR=0/1, WR=1/0/1, 8 bit data bus
	clrf	PORTA		; SRST=0 (reset), DMAEN=0 (no DMA/IRQ)
	movlw	0x68		; /LDOR=0, /WR=0, A=0, /CS=1: write FDD controller operations register, base mode
	movwf	PORTC
	movlw	0xF8		; A=0, all others are 1
	movwf	PORTC
	movlw	0xFF
	movwf	PORTA		; setting data bus to high state

	movlw	0xFF
	call	Wait_ms		; wait 255 ms

	call	ReadMainStatRegFDC

	movlw	0x04		; 04???		; release reset, DMAEN=0 (disabled)
	movwf	PORTA
	movwf	DORVAL
	movlw	0x68		; /LDOR=0, /WR=0, A=0, /CS=1: write FDD controller operations register, base mode
	movwf	PORTC
	movlw	0xF8
	movwf	PORTC
	movlw	0xFF
	movwf	PORTA
; --- RESET COMPLETE ---

	movlw	0xFF
	call	Wait_ms		; wait 255 ms

	call	ReadMainStatRegFDC
	call	SenseInterruptCmd

	movlw	0xFF
	movwf	TIMOUT1		; timeout: 255 loops
CheckAgain:
	call	ReadMainStatRegFDC	; read to get value and to get FDC stay in Base Mode
	btfsc	MAINSTAT,4
	goto	NotReady			; go there if bit 4 = 1 (command is executing)
	btfss	MAINSTAT,7
	goto	NotReady			; go there if bit 7 = 0 (FDC is not ready for data exchange)
	btfss	MAINSTAT,6
	goto	ReadyToWrite		; go there if bit 6 = 0 (ok to write to FDC)
	; FDC has something to say here
	movlw	0xFF
	movwf	PORTA		; port A - high level
	movlw	0xD1		; /FDDCS(0)+/RDI(0)+A0(1) = read data register
	movwf	PORTC
	movlw	0xF8
	movwf	PORTC
NotReady:
	decfsz	TIMOUT1,1
	goto	CheckAgain
	goto	SoftReset		; reset again after timeout

ReadyToWrite:
; --- Set up data rate - writing to CR

	call	SetSpeeds
	return


; ------------------------ Initialize drive ------------------------
; in:	CURDRV = current drive
; out:	RETERR = set if error occured at any of the init steps
InitDrive:
	; Turn motor on
	call	MotorOnSelect
	call	SetSpeeds

	; recalibrating drive
	call	Recalibrate
	btfsc	FDCERR,0
	goto	RetError

	; Seek to track 79
	call	NewFDCRequest
	movlw	0x0F		; SEEK command
	call	AddToFDCQueue
	movf	CURDRV,W
	call	AddToFDCQueue
	movlw	0x4F		; new cylinder number: 79
	call	AddToFDCQueue
	call	RunCommand	; no result phase
	call	WaitSeekComplete
	btfsc	FDCERR,0
	goto	RetError

; --- Seek to track 0, drive 0
	call	NewFDCRequest
	movlw	0x0F		; SEEK command
	call	AddToFDCQueue
	movf	CURDRV,W
	call	AddToFDCQueue
	movlw	0x00		; new cylinder number: 00
	call	AddToFDCQueue
	call	RunCommand	; no result phase
	call	WaitSeekComplete
	btfsc	FDCERR,0
	goto	RetError
	bcf		STATUS,C
	return

RetError:
	call	MotorOff	; start motor off countdown for current drive
	bsf		STATUS,C
	return
; ************************ Initialize drive ************************

; recalibrates CURDRV
; returns FDCERR!=0 on error
Recalibrate:
	; Recalibrate drive
	call	NewFDCRequest
	movlw	0x07		; RECALIBRATE command
	call	AddToFDCQueue
	movf	CURDRV,W
	call	AddToFDCQueue
	call	RunCommand
	call	WaitSeekComplete
	btfsc	FDCERR,0
	return

	; now we issue "SENSE DRIVE STATUS" command, and if track 0 signal (bit 4 of ST3)
	; is not set, drive is not present or defective
	call	NewFDCRequest			; clear MC FDC queue
	movlw	0x04
	call	AddToFDCQueue			; Sense drive status FDC command
	movf	CURDRV,W
	call	AddToFDCQueue
	call	RunCommand

	btfsc	FDCERR,0
	return

	movlw	0xFF
	btfss	0x60,4			; bit 4 of ST3, "TRACK 0", is not set
	movwf	FDCERR
	return

; --------------- Run command --------------------
; in:	none
; out:	status bytes in bank 0 starting offset 0x40,
;		FDCRUNPOS=number of results read, FDCERR<>0 if error
RunCommandWithData:
	movlw	0xFF
	movwf	CMDRUN		; set "command is running" flag so that host receive status "busy"
	call	MotorOnSelect
	clrf	FDCERR
	call	CommandPhase
	call	ExecutePhase
	call	ResultPhase
	call	MotorOff
	clrf	CMDRUN
	return
; *************** Run command ********************

; --------------- Execute phase --------------------
ExecutePhase:
	clrf	TIMOUT2
	clrf	TIMOUT1		; long timeout = 0x10000 loops

	; wait for execution phase start
ExecWait:
	movlw	4
	call	Wait_us			; wait 4 us
	call	ReadMainStatRegFDC
	btfsc	MAINSTAT,5
	goto	ContinueExec

	; we are here if execution phase did not start (yet)
	; checking for RQM bit - if it's 1 but EXM is still 0, then
	; skip execution phase (probably error occured)
;	btfsc	MAINSTAT,7
;	goto	ExecError

	; and wait in loop
	decfsz	TIMOUT1,1
	goto	ExecWait
	decfsz	TIMOUT2,1
	goto	ExecWait
ExecError:
	movlw	0xFF
	movwf	FDCERR
	return

ContinueExec:
	clrf	RECLENL
	clrf	RECLENH			; byte count in the record

	clrf	TIMOUT2
	clrf	TIMOUT1			; long timeout = 0x10000 loops

	movlw	0x01			; bank 4
	movwf	FSR0H	
	movlw	0x20			; start buffer pointer
	movwf	FSR0L

ExecuteSense:
	movlw	4
	call	Wait_us			; wait 4 us
	call	ReadMainStatRegFDC
	btfss	MAINSTAT,5		; EXM bit, equal to 0 when execution is complete
	return					; execute phase complete (with FDCERR preserved)

	btfsc	MAINSTAT,7
	goto	ExecRead

	decfsz	TIMOUT1,1
	goto	ExecuteSense
	decfsz	TIMOUT2,1
	goto	ExecuteSense
	goto	ExecError		; timeout error

ExecRead:
	btfss	MAINSTAT,6
	goto	ExecWrite		; data write requested unexpectedly

	; data read from the FDC
	call	ReadDataRegFDC
	movwi	FSR0++			; write data to buffer and increase FSR0
	goto	NextCount

ExecWrite:
	; data write to the FDC
	moviw	FSR0++			; read data from buffer and increase FSR0
	call	WriteDataRegFDC

NextCount:
	incfsz	RECLENL,1
	goto	NextPtr
	incf	RECLENH,1		; increase byte count

NextPtr:
	movf	FSR0,W
	andlw	0x7F			; mask all bits except bit 7
	sublw	0x60			; end of buffer chunk = 0x5F
	btfss	STATUS,Z
	goto	ExecuteSense	; still within current buffer chunk

	movlw	0x40
	addwf	FSR0L,1
	movlw	0x00
	addwfc	FSR0H,1			; increment 16-bit pointer by 0x40 (start of next chunk)
	goto	ExecuteSense
; *************** Execute phase ********************

; --------------- Run command --------------------
; in:	none
; out:	status bytes in bank 0 starting offset 0x40,
;		FDCRUNPOS=number of results read, FDCERR<>0 if error
RunCommand:
	clrf	FDCERR
	movlw	0xFF
	movwf	CMDRUN		; set "command is running" flag so that host receive status "busy"
	call	MotorOnSelect
	call	CommandPhase
	movlw	0x20
	call	Wait_us			; wait 32 microseconds
	call	ResultPhase
	call	MotorOff
	clrf	CMDRUN
	return
; *************** Run command ********************

; --------------- Reads data byte from the FDD controller --------------------
; in:	none
; out	data byte in W
ResultPhase:
	clrf	FSR0H		; indirect addressing bank 0
	movlw	0x60
	movwf	FSR0L		; low byte of the pointer
	clrf	FDCRUNPOS

	movlw	0xFF		; long timout! ; 0x08
	movwf	TIMOUT2
	clrf	TIMOUT1		; timeout = 0x800 loops
ReadLoop:
	call	ReadMainStatRegFDC
	btfsc	MAINSTAT,7	; 0x80, RQM
	goto	DataRegReady
	movlw	0x30
	call	Wait_us

	decfsz	TIMOUT1,1
	goto	ReadLoop
	decfsz	TIMOUT2,1
	goto	ReadLoop

	; timeout - exit
ReadErrorExit:
	movlw	0xFF
	movwf	FDCERR
	return

CheckCommandEnd:
	btfss	MAINSTAT,4	; 0x10, CB
	return				; result phase complete - return without error
	goto	ReadErrorExit

DataRegReady:
	btfss	MAINSTAT,6	; DIO - should be 1, i.e. data ready for reading
	goto	CheckCommandEnd

	call	ReadDataRegFDC
	movwi	FSR0++		; write byte to the indirectly addressed location

	incf	FDCRUNPOS,1	; increment current pointer to the queue, increase pointer
	movf	FDCRUNPOS,W
	sublw	0x10
	btfss	STATUS,Z
	goto	ReadLoop
	goto	ReadErrorExit		; result buffer overrun (max 16 result bytes)
; *************** Reads data byte from the FDD controller ********************

; ----------------------- COMMAND PHASE ------------------------
; Writes queue contents to the FDC
CommandPhase:
	clrf	FSR0H		; indirect addressing bank 0
	movlw	0x50
	movwf	FSR0L		; low byte of pointer to the buffer
	clrf	FDCRUNPOS

	clrf	TIMOUT2
	clrf	TIMOUT1		; timeout = 0x800 loops

WriteLoop:
	call	ReadMainStatRegFDC
	btfsc	MAINSTAT,7	; 0x80, RQM
	goto	WriteRQMReady
CPhWait:
	movlw	0x30
	call	Wait_us
	decfsz	TIMOUT1,1
	goto	WriteLoop
	decfsz	TIMOUT2,1
	goto	WriteLoop
	; timeout - exit

WriteErrorExit:
	movlw	0xFF
	movwf	FDCERR
	return

WriteRQMReady:
	btfsc	MAINSTAT,6
	goto	CPhWait			;WriteErrorExit		; DIO=1 means controller has unexpectedly something to say
	moviw	FSR0++				; read byte from the indirectly addressed location and increase pointer
	call	WriteDataRegFDC

	incf	FDCRUNPOS,1			; increment current pointer to the queue
	movf	FDCRUNPOS,W
	subwf	FDCQPOS,W
	btfss	STATUS,Z
	goto	WriteLoop

	movlw	0x20
	call	Wait_us
	return						; FDCERR is preserved
; *********************** COMMAND PHASE ************************

; ------------------- Reads FDC data register -----------------------
; returns data byte in W
ReadDataRegFDC:
	movlw	0xFF
	movwf	PORTA		; port A - high level
	movlw	0xD1		; /FDDCS+/RDI+A0(1) = read data register
	movwf	PORTC
	BANKSEL	TRISA
	movlw	0xFF
	movwf	TRISA
	movlb	0
	movf	PORTA,W
	movwf	DATAVAL
	BANKSEL	TRISA
	clrf	TRISA
	movlb	0
	movlw	0xF8		; A=0, all others are 1
	movwf	PORTC
	movf	DATAVAL,W
	return
; ******************* Reads FDC data register ***********************

; ------------------- Writes to FDC data register -----------------------
; in:	W = data byte to write
WriteDataRegFDC:
	movwf	PORTA
	movlw	0xC9		; /FDDCS+/WRI+A0(1) = write data register
	movwf	PORTC
	movlw	0xF8		; A=0, all others are 1
	movwf	PORTC
	nop					; 125 ns delay to for FDC to latch and process byte
	return
; ******************* Writes to FDC data register ***********************

; --------------- Reads FDC master status register ----------------------
; returns status byte in W
ReadMainStatRegFDC:
;	movlw	0x18		; 24 us
;	call	Wait_us		; wait 24 us as per chip specification
	movlw	0xFF
	movwf	PORTA		; port A - high level
	movlw	0xD0		; /FDDCS+/RDI+A0 = read master register
	movwf	PORTC
	BANKSEL	TRISA
	movlw	0xFF
	movwf	TRISA
	movlb	0
	movf	PORTA,W
	movwf	MAINSTAT
	BANKSEL	TRISA
	clrf	TRISA
	movlb	0
	movlw	0xF8		; A=0, all others are 1
	movwf	PORTC
	movf	MAINSTAT,W
	return
; *************** Reads FDC master status register **********************

; -------------- Adds byte to the FDD command queue ---------------------
AddToFDCQueue:
	movwf	QTEMP
	clrf	FSR0H		; indirect addressing bank 0
	movf	FDCQPOS,W
	addlw	0x50
	movwf	FSR0L		; visible in all banks
	movf	QTEMP,W
	movwi	[FSR0]		; write to the indirectly addressed location
	incf	FDCQPOS,1
	return
; ************** Adds byte to the FDD command queue *********************

; ---------------- Clears FDD command/data queue ------------------------
NewFDCRequest:
	clrf	FDCQPOS
	return
; **************** Clears FDD command/data queue ************************

; ------------------------ SENSE INTERRUPT -------------------------------
SenseInterruptCmd:
	movlw	0x0A
	call	Wait_ms					; wait 10 ms
	call	NewFDCRequest
	movlw	0x08		; SENSE INTERRUPT command
	call	AddToFDCQueue
	call	RunCommand
	return
; ************************ SENSE INTERRUPT *******************************

; ---------------- Wait millisecond, number of milliseconds in W -----------------
; W is preserved
Wait_ms:
	movwf	DELAY3
	clrf	DELAY1
Wait_ms2:
	movlw	0x0A
	movwf	DELAY2
Wait_ms1:
	decfsz	DELAY1,1
	goto	Wait_ms1
	decfsz	DELAY2,1
	goto	Wait_ms1
	decfsz	DELAY3,1
	goto	Wait_ms2
	return	
; **************** Wait millisecond, number of milliseconds in W *****************

; ---------------- Wait microsecond, number of microseconds in W -----------------
Wait_us:
	movwf	DELAY1
Wait_us1:
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	decfsz	DELAY1,1
	goto	Wait_us1
	return
; **************** Wait microsecond, number of microseconds in W *****************

; ------------------ Turn motor on and clear turnoff counter ----------------
; in:	CURDRV = current drive #
MotorOnSelect:
;;;	bcf		INTCON,GIE	; disable interrupts
	clrf	TMP
	movf	DORVAL,W
	btfss	CURDRV,0
	goto	Drv0

	btfsc	DORVAL,5	; bit of drive motor running
	incf	TMP,1		; TMP to change to 1	

	clrf	MOTBTO		; no drive B timeout - keep it running
	andlw	0xFC		; mask drive select bits
	iorlw	0x21		; turn drive B motor on and select drive B
	goto	WriteDOR

Drv0:
	btfsc	DORVAL,4	; bit of drive motor running
	incf	TMP,1		; TMP to change to 1	

	clrf	MOTATO		; no drive A timeout - keep it running
	andlw	0xFC		; mask drive select bits
	iorlw	0x10		; turn drive A motor on and select drive A

WriteDOR:
	movwf	DORVAL		; save current DOR value
	movwf	PORTA
	movlw	0x68		; /LDOR=0, /WR=0, A=0, /CS=1: write FDD controller operations register, base mode
	movwf	PORTC
	movlw	0xF8
	movwf	PORTC

;;;	bsf		INTCON,GIE	; enable interrupts

	btfsc	TMP,0
	return				; skip drive spin up delay if TMP<>0 (selected drive motor was already running)

	movlw	0xFF
	movwf	PORTA
	call	Wait_ms
	movlw	0xFF
	call	Wait_ms		; wait 255*2=510 ms
	movlw	0xFF
	call	Wait_ms
	movlw	0xFF
	call	Wait_ms
	return
; ****************** Turn motor on and clear turoff counter ****************

; ------------------ Set up counter to turn motor off ----------------
; in:	CURDRV = current drive #
MotorOff:
	btfss	CURDRV,0
	goto	MotorCountDownDrv0

	movlw	0xFF
	movwf	MOTBTO		; drive B timeout set
	return

MotorCountDownDrv0:
	movlw	0xFF
	movwf	MOTATO		; drive A timeout set
	return
; ****************** Set up counter to turn motor off ****************

; ------------------ Wait for SEEK operation completes ----------------
WaitSeekComplete:
	call	SenseInterruptCmd
	btfss	0x60,5
	goto	WaitSeekComplete	; loop while SE bit is 0 (seek is not complete)
	return
; ****************** Wait for SEEK operation completes ****************


; ------------------ Clear data buffer  ---------------------
; in:	none
ClearBufForFormat:
	clrf	RECLENL
	movlw	0x02
	movwf	RECLENH
	movlw	0x01
	movwf	FSR0H
	movlw	0x20
	movwf	FSR0L

ClearFillLoop:
	clrw
	movwi	FSR0++

	decfsz	RECLENL,1
	goto	NextClearBuf
	decf	RECLENH,1		; increase byte count
	btfsc	STATUS,Z
	return

NextClearBuf:
	movf	FSR0,W
	andlw	0x7F			; mask all bits except bit 7
	sublw	0x60			; end of buffer chunk = 0x5F
	btfss	STATUS,Z
	goto	ClearFillLoop	; still within current buffer chunk

	movlw	0x40
	addwf	FSR0L,1
	movlw	0x00
	addwfc	FSR0H,1			; increment 16-bit pointer by 0x40 (start of next chunk)
	goto	ClearFillLoop
; ****************** Clear data buffer  *********************

; ------------------ Fill data buffer with data for format 1.44MiB diskette ------------
; in:	W = cylinder to format, 18 sectors per track
; (!!!) Important: sector interleaving is NOT implemented
FillBufForFormat:
	movwf	DATAVAL
	clrf	RECLENL
	clrf	RECLENH
	movlw	0x01
	movwf	TMP			; sector number
	movwf	FSR0H
	movlw	0x20
	movwf	FSR0L

FormatFillCycle:
	movf	DATAVAL,W		; cylinder # (track)
	movwi	FSR0++
	lsrf	0x51,W			; second byte in the FORMAT command sequence
	lsrf	WREG,W
	andlw	0x01
	movwi	FSR0++			; head
	movf	TMP,W
	movwi	FSR0++			; sector#
	movlw	0x02
	movwi	FSR0++			; sector size (0x200)
	incf	TMP,1

	movf	TMP,W
	sublw	0x13			; let routine fill all 18 sector space
	btfsc	STATUS,Z		; even if we format 9 sectors per track disk
	return

	movf	FSR0,W
	andlw	0x7F			; mask all bits except bit 7
	sublw	0x60			; end of buffer chunk = 0x5F
	btfss	STATUS,Z
	goto	FormatFillCycle	; still within current buffer chunk

	movlw	0x40
	addwf	FSR0L,1
	movlw	0x00
	addwfc	FSR0H,1			; increment 16-bit pointer by 0x40 (start of next chunk)
	goto	FormatFillCycle
; ****************** Fill data buffer with data for format 1.44MiB diskette ************

;============================= Next subroutines are host/MC access code =========================
; ---------------- Waits for host write --------------------
; out:	data read from the host (W), C=1 if A0 was 1
; waits for write in loop. Interrupts are enabled. If write is missed for some reason,
; this sub will continue listening for write, and next host's read will return 0xFF, thus
; procedure to access MC is: host writes to port, then reads from port. If read value is
; 0xFF, then write again in loop (with timeout). If it reads 0xAA, then perform real data
; write or read. During real write/read, MC interruprs are disabled.
WaitForHostInitWrite:
	movf	PORTE,W
	movf	PORTD,1		; port D value into the LATD
	btfss	WREG,3
	goto	WaitForHostInitWrite

DecodeHostRetVal:
	rrf		WREG,W		; shift address bit (bit 0) to C
	BANKSEL	LATD
	movf	LATD,W
	BANKSEL	PORTD
	return

RetStatAndWaitForHostWrite:
; in:	port D in input mode
; out:	data read from host (W)
; waits for second host write byte. First, it disables interrupts, and then returns first byte
; status 0xAA, and on the next CPU write gets byte from from data bus
	call	ReturnStatusHost
	; now wait for host write
	call	WaitForHostInitWrite
	return

ReturnStatusHost:
; returns 0xAA to the CPU read. During and after exit of this sub interrupts are disabled, thus CPU should
; read this status byte shortly with specific delay to let MC run its interrupt procedure
	movf	COUVAL,W
ReturnByteToHost:		;HostRead:
	movwf	RESULTN
	movlw	0xFF
	movwf	PORTD
	BANKSEL	TRISD
	clrf	TRISD
	BANKSEL	PORTD
	movf	RESULTN,W

WaitForHostReadLoop:
	btfss	PORTE,2
	goto	WaitForHostReadLoop

	movwf	PORTD
	movlw	0xFF

WaitForHostReadLoop1:
	btfsc	PORTE,2
	goto	WaitForHostReadLoop1		; Waiting for release of /PICRD

	movwf	PORTD						; releasing data bus by setting all lines to logical 1

	BANKSEL	TRISD
	movwf	TRISD
	BANKSEL	PORTD						; data bus now is in input mode
	return

RetStatToHostAndWaitForHostRead:
; in:	data to write to host (W)
; returns 0xB6 to the host, then waits host strobe write, and then writes COUVAL, and then writes data byte
; out: returns flag C=0 if stobe write (after MC outputs 0xB6) was performed into port base+0 (i.e. command abort signal)
	movwf	DATAVAL
	call	HostSync				; if no error returns with port D in output mode
	btfss	STATUS,C
	return

	movf	DATAVAL,W
	movwf	PORTD					; putting DATAVAL to the bus

WaitForRDRelease2:
	btfss	PORTE,2
	goto	WaitForRDRelease2		; Waiting for activation of /PICRD (COUVAL was read by the host)

WaitForRDRelease3:
	btfsc	PORTE,2
	goto	WaitForRDRelease3		; Waiting for release of /PICRD

	movlw	0xFF
	movwf	PORTD					; setting all data bits in port D output to 1
	BANKSEL	TRISD
	movwf	TRISD
	BANKSEL	PORTD					; data bus (port D) now is in input mode
	bsf		STATUS,C				; set bit = assuming it read from base port +1
	return

RetStatToHostAndWaitForHostWrite:
; out:	data received from host (W)
; returns 0xB6 to the host, then waits host strobe write, and then writes COUVAL, and then reads data byte
	call	HostSync
	btfss	STATUS,C
	return

	BANKSEL	TRISD
	comf	TRISD,1					; 0x00 -> 0xFF
	BANKSEL	PORTD					; data bus (port D) now is in input mode

WaitForHostWrt:
	movf	PORTE,W
	movf	PORTD,1					; port D value into the LATD
	btfss	WREG,3
	goto	WaitForHostWrt

	rrf		WREG,W		; shift address bit (bit 0) to C
	BANKSEL	LATD
	movf	LATD,W
	BANKSEL	PORTD
	return							; C flag is returned

HostSync:
; Synchronization of data exchange with with the host
; (return 0xb6, wait for host write, then return COUVAL)
; if no error returns to caller with port D in output mode!
; out: C=1 if port base+0 is detected during handshake process (0xB6)
	movlw	0xFF
	movwf	PORTD					; putting 0xFF to port D

	BANKSEL	TRISD
	clrf	TRISD
	BANKSEL	PORTD

	movlw	0xB6
	movwf	PORTD					; putting 0xB6 to port D - should happen as close to next host write sense as possible

	call	WaitForHostInitWrite	; waiting for host strobe write

	btfsc	STATUS,C
	goto	RSTH0xA5

	BANKSEL	TRISD
	comf	TRISD,1					; 0x00 -> 0xFF
	BANKSEL	PORTD					; data bus (port D) now is in input mode
	return							; C flag = 0 is preserved since last bit test operator

RSTH0xA5:
	movf	COUVAL,W
	movwf	PORTD					; putting status byte onto data bus

WaitForHostLoop0:
	btfss	PORTE,2
	goto	WaitForHostLoop0		; Waiting for activation of /PICRD

WaitForHostLoop1:
	btfsc	PORTE,2
	goto	WaitForHostLoop1		; Waiting for release of /PICRD (COUVAL was read by the host)

	bsf		STATUS,C				; emulating write to base port +1
	return
;============================= End of "Next subroutines are host/MC access code" =========================

CheckSingleDrive:
; in:	drive number requested
	movwf	CURHOSTDRV
	movwf	CURDRV
	movf	DRINUM,W
	sublw	0x01			; is there only one drive?
	btfsc	STATUS,Z
	clrf	CURDRV			; current drive = drive 0 (while it can be phantom drive B: for host)
	return

int_handler:
	movlb	0

	; if MOTxTO is not 0, decrement it and when it is zero stop the motor but not deselect drive
	movf	MOTATO,1		; check Z flag
	btfsc	STATUS,Z
	goto	StopDrvB		; MOTATO=0, do not count it down

	decfsz	MOTATO,1
	goto	StopDrvB		; counter did not reach 0 for drive A

	movf	DORVAL,W
	andlw	0xEF			; bit 4 reset
	call	WriteDOR

StopDrvB:
	movf	MOTBTO,1		; check Z flag
	btfsc	STATUS,Z
	goto	RetInt			; MOTBTO=0, do not count it down

	decfsz	MOTBTO,1
	goto	RetInt			; counter did not reach 0 for drive B

	movf	DORVAL,W
	andlw	0xDF			; bit 5 reset
	call	WriteDOR

RetInt:
	clrf	TMR1L
	clrf	TMR1H
	clrf	PIR1			; located in bank 0
	retfie
	END